use std::borrow::Cow;

use quick_xml::{events::attributes::Attribute, name::QName};

use self::{
    primitives::{
        blend::{Blend, BlendMode},
        color_matrix::{ColorMatrix, ColorMatrixType},
        component_transfer::{ComponentTransfer, TransferFn},
        composite::{Composite, CompositeOperator},
        gaussian_blur::GaussianBlur,
        offset::Offset,
        FePrimitive,
    },
    standard_input::StandardInput,
};

use super::length::{Coordinate, Length};

pub mod primitives;
pub mod standard_input;

#[derive(Debug)]
pub enum Node {
    StdInput(StandardInput),
    Primitive {
        primitive: FePrimitive,
        common_attrs: CommonAttrs,
    },
}

impl Default for Node {
    fn default() -> Self {
        Self::StdInput(StandardInput::SourceGraphic)
    }
}

#[derive(Default, Debug, Clone, Copy)]
pub(crate) struct CommonAttrs {
    pub x: Coordinate,
    pub y: Coordinate,
    pub width: Length,
    pub height: Length,
}

impl From<CommonAttrs> for Vec<Attribute<'_>> {
    fn from(val: CommonAttrs) -> Self {
        let mut r = Vec::new();
        if !val.x.is_zero() {
            r.push(Attribute {
                key: QName(b"x"),
                value: Cow::from(val.x.to_string().into_bytes()),
            });
        }

        if !val.y.is_zero() {
            r.push(Attribute {
                key: QName(b"y"),
                value: Cow::from(val.y.to_string().into_bytes()),
            });
        }

        if !val.width.is_zero() {
            r.push(Attribute {
                key: QName(b"width"),
                value: Cow::from(val.width.to_string().into_bytes()),
            });
        }

        if !val.height.is_zero() {
            r.push(Attribute {
                key: QName(b"height"),
                value: Cow::from(val.height.to_string().into_bytes()),
            });
        }

        r
    }
}

impl Node {
    pub fn simple(el: FePrimitive) -> Node {
        Node::Primitive {
            primitive: el,
            common_attrs: CommonAttrs::default(),
        }
    }

    pub fn primitive(&self) -> Option<(&FePrimitive, &CommonAttrs)> {
        if let Node::Primitive {
            primitive,
            common_attrs,
        } = self
        {
            Some((primitive, common_attrs))
        } else {
            None
        }
    }

    pub fn input_count(&self) -> u8 {
        match self {
            Node::Primitive {
                primitive:
                    FePrimitive::ColorMatrix(_)
                    | FePrimitive::ComponentTransfer(_)
                    | FePrimitive::ConvolveMatrix(_)
                    | FePrimitive::DiffuseLighting(_)
                    | FePrimitive::GaussianBlur(_)
                    | FePrimitive::Morphology(_)
                    | FePrimitive::Offset(_)
                    | FePrimitive::SpecularLighting(_)
                    | FePrimitive::Tile(_),
                ..
            } => 1,

            Node::Primitive {
                primitive:
                    FePrimitive::Composite(_) | FePrimitive::Blend(_) | FePrimitive::DisplacementMap(_),
                ..
            } => 2,

            Node::StdInput(_)
            | Node::Primitive {
                primitive:
                    FePrimitive::Flood(_) | FePrimitive::Image(_) | FePrimitive::Turbulence(_),
                ..
            } => 0,
            Node::Primitive {
                primitive: FePrimitive::Merge(_),
                ..
            } => todo!(),
        }
    }

    pub fn blend(mode: BlendMode) -> Self {
        Self::simple(FePrimitive::Blend(Blend::new(mode)))
    }

    pub fn color_matrix(cm_type: ColorMatrixType) -> Self {
        Self::simple(FePrimitive::ColorMatrix(ColorMatrix::new(cm_type)))
    }

    pub fn composite(op: CompositeOperator) -> Self {
        Self::simple(FePrimitive::Composite(Composite::new(op)))
    }

    pub fn composite_arithmetic(k1: f32, k2: f32, k3: f32, k4: f32) -> Self {
        Self::composite(CompositeOperator::Arithmetic { k1, k2, k3, k4 })
    }

    pub fn gaussian_blur(v: u16) -> Self {
        Self::simple(FePrimitive::GaussianBlur(GaussianBlur::single(v)))
    }

    pub fn gaussian_blur_xy(x: u16, y: u16) -> Self {
        Self::simple(FePrimitive::GaussianBlur(GaussianBlur::with_xy(x, y)))
    }

    pub fn offset(dx: f32, dy: f32) -> Self {
        Self::simple(FePrimitive::Offset(Offset::new(dx, dy)))
    }

    pub fn component_transfer_rgba(
        r: TransferFn,
        g: TransferFn,
        b: TransferFn,
        a: TransferFn,
    ) -> Self {
        Self::simple(FePrimitive::ComponentTransfer(ComponentTransfer {
            func_r: r,
            func_g: g,
            func_b: b,
            func_a: a,
        }))
    }
}
