use std::borrow::Cow;

use quick_xml::{
    events::attributes::{Attr, Attribute},
    name::QName,
};

use super::WriteElement;

/// [feTurbulence](https://www.w3.org/TR/SVG11/filters.html#feTurbulenceElement)
#[derive(Debug)]
pub struct Turbulence {
    base_frequency: (f32, f32),
    num_octaves: u16,
    seed: u32,
    stitch_tiles: StitchTiles,
    // attr name: type
    noise_type: NoiseType,
}

impl WriteElement for Turbulence {
    fn attrs(&self) -> Vec<quick_xml::events::attributes::Attribute> {
        let mut r = Vec::from([Attribute {
            key: QName(b"baseFrequency"),
            value: Cow::from(
                format!("{} {}", self.base_frequency.0, self.base_frequency.1).into_bytes(),
            ),
        }]);

        if self.num_octaves != 1 {
            r.push(Attribute {
                key: QName(b"numOctaves"),
                value: Cow::from(format!("{}", self.num_octaves).into_bytes()),
            });
        }

        if self.seed != 0 {
            r.push(Attribute {
                key: QName(b"seed"),
                value: Cow::from(self.seed.to_string().into_bytes()),
            });
        }

        if self.stitch_tiles != StitchTiles::NoStitch {
            r.push(Attribute {
                key: QName(b"stitchTiles"),
                value: Cow::from(b"stitch".to_vec()),
            });
        }

        if self.noise_type != NoiseType::Turbulence {
            r.push(Attribute {
                key: QName(b"type"),
                value: Cow::from(b"fractalNoise".to_vec()),
            });
        }

        r
    }

    fn tag_name(&self) -> &'static str {
        "feTurbulence"
    }
}

#[derive(Debug, PartialEq, Eq, Clone, Copy)]
pub enum StitchTiles {
    Stitch,
    NoStitch,
}
#[derive(Debug, PartialEq, Eq, Clone, Copy)]
pub enum NoiseType {
    Turbulence,
    FractalNoise,
}
